"""Python API for podman interface in maple"""

import subprocess


def login():
    """
    Login to podman account
    """
    subprocess.run("podman login docker.io", shell=True, check=True)


def prune():
    """
    Prune system
    """
    subprocess.run("rm -f -v $maple_home/context/Dockerfile*", shell=True, check=True)
    subprocess.run("podman system prune -a", shell=True, check=True)



"""
Command Line Interface (CLI) for image management.
"""

import os
import click
import toml

from ..backend import Backend
from . import maple


# CLI group
#
@maple.group(name="image")
def image():
    """
    Image management, type maple image --help for more info
    """
    pass


# Build a local image using a base image
# The modulermation supplied from Maplefile
@image.command(name="build")
@click.argument("image", default="None")
@click.option("--base", "-B", default="None", help="base image name")
@click.option("--as-root", is_flag=True, help="flag to build image as root")
@click.option("--options", "-o", default="", help="run options")
def build(image, base, as_root, options):
    """
    Builds a local image from a base image
    """
    if base != "None":
        os.environ["maple_base"] = str(base)
    if image != "None":
        os.environ["maple_image"] = str(image)

    Maplefile = os.path.exists("Maplefile")

    cmd_list = []
    env_list = []
    if Maplefile:
        if "build" in toml.load("Maplefile"):
            cmd_list = toml.load("Maplefile")["build"]
        if "environ" in toml.load("Maplefile"):
            env_list = toml.load("Maplefile")["environ"]

    Backend().image.build(as_root, cmd_list, env_list, options)


# Pull base image from a remote registry
@maple.command(name="pull")
@click.argument("target", required=True)
@click.argument("base", required=True)
def pull(target, base):
    """
    Pull image from remote registry
    """
    Backend().image.pull(target, base)


# Push image to remote registry
# Note will require 'maple login' if credentials are required
#
@maple.command(name="push")
@click.argument("base", required=True)
@click.argument("target", required=True)
def push(base, target):
    """
    Push image to remote registry
    """
    Backend().image.push(base, target)


# Tag an image from base
@maple.command(name="tag")
@click.argument("base", required=True)
@click.argument("target", required=True)
def tag(base, target):
    """
    Tag a target image from base image
    """
    Backend().image.tag(base, target)


# List all images
#
@image.command("list")
def list():
    """
    List all images on system
    """
    Backend().image.list()


# Squash and prune layers
#
@image.command("squash")
@click.argument("image", default="None")
def squash(image):
    """
    Squash and remove layers from local image, reduces size of the image
    """
    if image != "None":
        os.environ["maple_image"] = str(image)
    Backend().image.squash()


# Scan all the images
@maple.command("scan")
@click.argument("image", required=True)
def scan(image):
    """
    Scan local images, accepts multiple arguments
    """
    Backend().image.scan(image)


# Clean all local images and containers
#
@image.command("delete")
@click.argument("images", nargs=-1)
def delete(images):
    """
    Delete local images, accepts multiple arguments
    """
    if not images:
        images = ["None"]

    for img in images:
        if img != "None":
            os.environ["maple_image"] = str(img)
        Backend().image.delete()



"""
Command Line Interface (CLI) for system configuration management.
"""
import os
import subprocess
import click

from ..backend import Backend
from . import maple


# Setup environment
#
@maple.command(name="config")
@click.argument("maplefile")
def config(maplefile):
    """
    Setup a configuration file from user provided path
    """
    if maplefile == "Maplefile":
        print("Cannot link a file to itself. Operation not permitted. ABORTING")
        raise ValueError()

    if os.path.exists("Maplefile"):
        print(f"Maplefile already exists in current directory")
        overwrite = input("Overwrite? Y/n ")

        if overwrite == "y" or overwrite == "Y":
            print("OVERWRITING")
            subprocess.run(
                f"rm Maplefile && ln -s {maplefile} Maplefile", shell=True, check=True
            )
        else:
            print("SKIPPING")
    else:
        subprocess.run(f"ln -s {maplefile} Maplefile", shell=True, check=True)


# Login to remote registry
#
@maple.command(name="login")
def login():
    """
    Login to container backend
    """
    Backend().system.login()


# Prune system
#
@maple.command("prune")
def prune():
    """
    Clear backend cache
    """
    Backend().system.prune()



"""Python API for docker interface in maple"""

import subprocess


def login():
    """
    Login to docker account
    """
    subprocess.run("docker login", shell=True, check=True)


def prune():
    """
    Prune system
    """
    subprocess.run("rm -f -v $maple_home/context/Dockerfile*", shell=True, check=True)
    subprocess.run("docker system prune -a", shell=True, check=True)



"""Python API for podman interface in maple"""

import os
import subprocess

from . import container


def build(as_root=False, cmd_list=None, env_list=None, options=""):
    """
    Builds a local image from remote image

    Arguments
    ---------
    as_root    : Build image as root (True/False)
    cmd_list   : Command list for build
    env_list   : List of persistent environment variables
    """
    if as_root:
        print("Rootless mode only with podman backend. ABORTING")
        raise ValueError()

    # Create a context directory
    subprocess.run(
        f'mkdir -pv {os.getenv("maple_home")}/context', shell=True, check=True
    )

    # Set Dockerfile for the build
    dockerfile_build = (
        os.getenv("maple_home") + "/context/Dockerfile." + os.getenv("maple_image")
    )

    # Select the base and user Dockerfile
    dockerfile_base = os.getenv("maple_dir") + "/resources/Dockerfile.base"

    # Populate Dockerfile for the build
    subprocess.run(
        f"cat {dockerfile_base} > {dockerfile_build}", shell=True, check=True
    )

    with open(f"{dockerfile_build}", "a") as dockerfile:  # append mode
        if env_list:
            for variable in env_list:
                dockerfile.write(f"\nENV {variable}\n")
        if cmd_list:
            for command in cmd_list:
                dockerfile.write(f"\nRUN {command}\n")

    print(f"Building on platform: {str(os.getenv('maple_platform'))}")

    # execute podman build
    subprocess.run(
        f"podman build {options} --platform $maple_platform -t $maple_image --no-cache \
                                 --volume $maple_source:$maple_target \
                                 --build-arg maple_workdir=$maple_target \
                                 --build-arg maple_base=$maple_base \
                                 --build-arg maple_user=$maple_user \
                                 --build-arg maple_uid=$maple_uid \
                                 --build-arg maple_gid=$maple_gid \
                                 --file={dockerfile_build} \
                                 $maple_home/context",
        shell=True,
        check=True,
    )

    # subprocess.run('rm $maple_home/context/Dockerfile.build', shell=True, check=True)


def pull(target, base):
    """
    Pull remote image

    Arguments
    ---------
    target : target image to pull into
    base   : base image in remote registry
    """
    subprocess.run(f"podman pull {base}", shell=True, check=True)
    subprocess.run(f"podman tag {base} {target}", shell=True, check=True)


def push(base, target):
    """
    Push local image to remote tag/image

    Arguments
    ---------
    base   : base image
    target : target image to push
    """
    subprocess.run(f"podman tag {base} {target}", shell=True, check=True)
    subprocess.run(f"podman push {target}", shell=True, check=True)


def tag(base, target):
    """
    Tag a target image from base image

    Arguments
    ---------
    base   : base image
    target : target image to push
    """
    subprocess.run(f"podman tag {base} {target}", shell=True, check=True)


def list():
    """
    List all images on system
    """
    subprocess.run("podman images", shell=True, check=True)


def squash():
    """
    Squash an image and remove layers
    """
    os.environ["maple_container"] = os.environ["maple_image"] + "_container"

    container.pour()
    subprocess.run(
        "podman export $maple_container > $maple_image.tar", shell=True, check=True
    )
    subprocess.run(
        "cat $maple_image.tar | podman import - $maple_image", shell=True, check=True
    )
    subprocess.run("rm $maple_image.tar", shell=True, check=True)
    container.rinse()


def scan(image):
    """
    Scan an image

    Arguments
    ---------
    image : image name

    """
    subprocess.run(f"podman scan {image}", shell=True, check=True)


def delete():
    """
    Delete an image
    """
    subprocess.run(
        "podman rmi $maple_image $(podman images --filter dangling=true -q --no-trunc)",
        shell=True,
        check=True,
    )



"""Python API for singularity interface in maple"""

import os
import subprocess


def build(as_root=False, cmd_list=None, env_list=None, options=""):
    """
    Builds a local image from remote image

    Arguments
    ---------
    as_root    : Build image as root (True/False)
    cmd_list   : List of build commands
    env_list   : List of persistent environment variables
    """
    if as_root:
        print("Rootless mode only with singularity backend. ABORTING")
        raise ValueError()

    # Create image directory
    subprocess.run(
        f'mkdir -pv {os.getenv("maple_home")}/images', shell=True, check=True
    )

    # Build image
    subprocess.run(
        "singularity build --sandbox $maple_image.sif $maple_base", shell=True, check=True
    )
    subprocess.run(
        "mv $maple_image.sif $maple_home/images/$maple_image.sif",
        shell=True,
        check=True,
    )


def pull(target, base):
    """
    Pull remote image

    Arguments
    ---------
    target : target image to pull into
    base   : base image in remote registry
    """
    subprocess.run(
        f"singularity pull $maple_home/images/{target}.sif {base}",
        shell=True,
        check=True,
    )


def push(base, target):
    """
    Push local image to remote tag/image

    Arguments
    ---------
    base   : base image
    target : target image to push
    """
    subprocess.run(
        f"singularity pull $maple_home/images/{base}.sif {target}",
        shell=True,
        check=True,
    )


def tag(base, target):
    """
    Tag a target image from base image

    Arguments
    ---------
    base   : base image
    target : target image to push
    """
    subprocess.run(
        f"cp $maple_home/images/{base}.sif $maple_home/images/{target}.sif",
        shell=True,
        check=True,
    )


def list():
    """
    List all images on system
    """
    subprocess.run("ls $maple_home/images/*.sif 2> /dev/null", shell=True, check=True)


def squash():
    """
    Squash an image and remove layers
    """
    print("[maple.image.squash] not available for singularity backend")


def scan(image):
    """
    Scan an image

    Arguments
    ---------
    image : image name
    """
    print("[maple.image.scan] not available for singularity backend")


def delete():
    """
    Delete an image
    """
    subprocess.run(
        "rm -f -v $maple_home/images/$maple_image.sif", shell=True, check=True
    )



"""Python API for docker interface in maple"""

import os
import subprocess
import random


def commit():
    """
    Commit changes from local container to local image
    """
    subprocess.run(
        "docker commit $maple_container $maple_image", shell=True, check=True
    )


def pour(options=""):
    """
    Pour local image in a container, opposite of maple rinse

    Arguments
    ---------
    options : string of options
    """
    process = subprocess.run(
        f"docker run --entrypoint '/bin/bash' {options} -dit \
                     --platform $maple_platform --name $maple_container \
                     --mount type=bind,source=$maple_source,target=$maple_target \
                     $maple_image",
        shell=True,
        check=True,
    )

    if process.returncode != 0:
        raise Exception("[maple] Error inside container")


def rinse(rinse_all=False):
    """
    Stop and remove the local container, opposite of maple pour

    Arguments
    ---------
    rinse_all : (True/False) flag to rinse all container
    """
    if rinse_all:
        subprocess.run("docker stop $(docker ps -aq)", shell=True, check=True)
        subprocess.run("docker rm $(docker ps -aq)", shell=True, check=True)
    else:
        subprocess.run("docker stop $maple_container", shell=True, check=True)
        subprocess.run("docker rm $maple_container", shell=True, check=True)


def shell():
    """
    Get shell access to the local container
    """
    subprocess.run(
        "docker exec -it --workdir $maple_target $maple_container bash",
        shell=True,
        check=True,
    )


def run(command, options=""):
    """
    Run and rinse the local container

    Arguments
    ---------
    command : command string
    options : run options
    """
    os.environ["maple_container"] = (
        os.getenv("maple_container") + "_" + str(random.randint(1111, 9999))
    )

    command = f'"{command}"'
    process = subprocess.run(
        f"docker run --entrypoint '/bin/bash' {options} \
                     --name $maple_container \
                     --mount type=bind,source=$maple_source,target=$maple_target \
                     --workdir $maple_target \
                     $maple_image -c {command}",
        shell=True,
        check=True,
    )

    rinse()

    if process.returncode != 0:
        raise Exception("[maple] Error inside container")


def execute(command):
    """
    Run local image in a container

    Arguments
    ---------
    command: string of command to execute
    """
    command = f'"{command}"'
    process = subprocess.run(
        f"docker exec ---entrypoint '/bin/bash' -workdir $maple_target $maple_container -c {command}",
        shell=True,
        check=True,
    )

    return process.returncode


def publish(cmd_list=None):
    """
    Publish container to an image

    Arguments
    ---------
    cmd_list: list of commands to publish
    """
    pour()

    result_list = []

    if cmd_list:
        for command in cmd_list:
            result_list.append(execute(command))

    commit()
    rinse()

    if not all(result == 0 for result in result_list):
        raise Exception("[maple] Error inside container")


def notebook(port="4321"):
    """
    Launch ipython notebook inside the container

    Arguments
    ---------
    image : image name
    port  : port id ('4321')

    """
    os.environ["maple_container"] = (
        os.getenv("maple_container") + "_" + str(random.randint(1111, 9999))
    )

    pour(options=f"-p {port}:{port}")
    result = execute(f"jupyter notebook --port={port} --no-browser --ip=0.0.0.0")
    rinse()

    if result != 0:
        raise Exception("[maple] Error inside container")


def list():
    """
    List all containers on system
    """
    subprocess.run("docker container ls -a", shell=True, check=True)



"""
Command Line Interface (CLI) for container management.
"""

import os
import click
import toml

from ..backend import Backend
from . import maple


# CLI group
#
@maple.group(name="container")
def container():
    """
    Container management, type maple container --help for more info
    """
    pass


# Commit changes to a container
# Saves changes to local container as an image, currently uses docker
#
@container.command(name="commit")
@click.option("--image", "-I", default="None", help="image to commit changes")
def commit(image):
    """
    Commit changes from a poured container to an image
    """
    if image != "None":
        os.environ["maple_image"] = str(image)
    Backend().container.commit()


# Enter the shell environment of a container
#
@container.command("shell")
def shell():
    """
    Get shell access inside a poured container
    """
    Backend().container.shell()


# Launch a notebook inside the container
#
@container.command("notebook")
@click.option("--image", "-I", default="None", help="image to launch notebook")
@click.option("--port", "-p", default="4321", help="port for notebook server")
def notebook(image, port):
    """
    Launch ipython notebook inside a container using an image
    """
    if image != "None":
        os.environ["maple_image"] = str(image)
    Backend().container.notebook(port)


# Publish a container to an image
#
@container.command("publish")
@click.option("--image", "-I", default="None", help="image to pour and publish")
def publish(image):
    """
    Publish a container to an image using default execute statements
    """
    if image != "None":
        os.environ["maple_image"] = str(image)

    Maplefile = os.path.exists("Maplefile")

    cmd_list = []
    if Maplefile:
        if "publish" in toml.load("Maplefile"):
            cmd_list = toml.load("Maplefile")["publish"]

    Backend().container.publish(cmd_list)


# Run a command inside a container and commit changes
@container.command("run")
@click.option("--image", "-I", default="None", help="image to run")
@click.option("--options", "-o", default="", help="run options")
@click.argument("command", default="echo Hello World!")
def run(image, options, command):
    """
    Pour a container to a run a command and then rinse it
    """
    if image != "None":
        os.environ["maple_image"] = str(image)
    Backend().container.run(command, options)


# Execute a command in a container
@container.command("execute")
@click.argument("command", default="echo Hello World!")
def execute(command):
    """
    Execute command in a poured container
    """
    result = Backend().container.execute(command)
    if result != 0:
        raise Exception("[maple] Error inside container")


# Pour an image in a local container to access interactive shell
# If maple_source or maple_traget are present then they will be mounted inside the containter.
# This is useful for mounting maple_source for development
#
@container.command("pour")
@click.option("--image", "-I", default="None", help="image to pour inside container")
def pour(image):
    """
    Pour an image into a container
    """
    if image != "None":
        os.environ["maple_image"] = str(image)
    Backend().container.pour()


# Rinse a local container
# Do this if the local container is not needed
#
@container.command("rinse")
@click.argument("containers", nargs=-1)
@click.option("--all", is_flag=True, help="rinse all containers")
def rinse(containers, all):
    """
    Stop and remove containers, accepts multiple arguments
    """
    if not containers:
        containers = ["None"]

    for ctr in containers:
        if ctr != "None":
            os.environ["maple_container"] = str(ctr)
        Backend().container.rinse(all)


# List all container
#
@container.command("list")
def list():
    """
    List all containers on system
    """
    Backend().container.list()



"""
Command Line Interface (CLI) for Maple. Reads Maplefile
and sets environment variables
"""

import os
import pwd
import click
import toml

# CLI group
#
@click.group(name="maple")
@click.option("--docker", is_flag=True, help="option for docker backend (default) ")
@click.option("--singularity", is_flag=True, help="option for singularity backend")
@click.option("--podman", is_flag=True, help="option for podman backend")
def maple(docker, singularity, podman):
    """
    CLI for using docker/singularity/podman containers for HPC applications
    """
    # Check if required environment variables are defined in the Maplefile
    # If not then assign default values

    os.environ["maple_uid"] = str(os.getuid())
    os.environ["maple_gid"] = str(os.getgid())
    os.environ["maple_user"] = pwd.getpwuid(os.getuid())[0]

    Maplefile = os.path.exists("Maplefile")

    if Maplefile:
        for key, value in toml.load("Maplefile").items():
            if key not in ["build", "publish", "environ"]:
                os.environ["maple_" + key] = str(value)

    if not os.getenv("maple_backend"):
        os.environ["maple_backend"] = "docker"

    if not os.getenv("maple_platform"):
        os.environ["maple_platform"] = "linux/amd64"

    if docker:
        os.environ["maple_backend"] = "docker"
    if singularity:
        os.environ["maple_backend"] = "singularity"
    if podman:
        os.environ["maple_backend"] = "podman"

    # Condition to check if target and source directories are defined in the Maplefile
    # assign default if they are not, and deal with execptions
    if not os.getenv("maple_target"):
        os.environ["maple_target"] = "/home/mount"
    if not os.getenv("maple_source"):
        os.environ["maple_source"] = os.getenv("PWD")



"""Python API for maple"""

import os
import pwd

from . import Maple
from ..backend import Backend


class Image(Maple):
    """
    Class to manage images

    Parameters
    ----------
    **attributes  : dictionary of user attributes
                    { 'base'      : remote image name,
                      'name'      : name of the image,
                      'backend'   : container backend - docker/singularity

    """

    def __init__(self, **attributes):
        """ """
        self._name = "ubuntu"
        self._base = "ubuntu:latest"
        self._backend = "docker"

        super().__init__(self.__dict__, attributes)

        # Set values for user and group
        self._uid = str(os.getuid())
        self._gid = str(os.getgid())
        self._user = pwd.getpwuid(os.getuid())[0]

        # Set container name for environment variables
        self._image = self._name

    @property
    def name(self):
        """
        Getter for image name
        """
        return self._name

    def build(self):
        """
        Builds a local image from base image
        """
        self.setenv()
        Backend().image.build()

    def squash(self):
        """
        Squash an the image
        """
        self.setenv()
        Backend().image.squash()

    def delete(self):
        """
        Delete the image
        """
        self.setenv()
        Backend().image.delete()



"""Initialize Maple"""
import os

os.environ["maple_dir"] = os.path.dirname(os.path.realpath(__file__))
os.environ["maple_home"] = os.environ["HOME"] + "/.local/maple"

from . import backend
from . import api
from . import cli
from . import resources



"""Build and installation script for PyMaple."""

# standard libraries
import re
from setuptools import setup, find_packages

# get long description from README
with open("README.rst", mode="r") as readme:
    long_description = readme.read()

with open("maple/__meta__.py", mode="r") as source:
    content = source.read().strip()
    metadata = {
        key: re.search(key + r'\s*=\s*[\'"]([^\'"]*)[\'"]', content).group(1)
        for key in [
            "__pkgname__",
            "__version__",
            "__authors__",
            "__license__",
            "__description__",
        ]
    }

# core dependencies - click, docker, singularity
DEPENDENCIES = ["click", "toml"]

setup(
    name=metadata["__pkgname__"],
    version=metadata["__version__"],
    author=metadata["__authors__"],
    description=metadata["__description__"],
    license=metadata["__license__"],
    packages=find_packages(where="./"),
    package_dir={"": "./"},
    package_data={
        "": [
            "resources/Dockerfile.base",
            "resources/Dockerfile.root",
            "resources/Dockerfile.user",
        ]
    },
    scripts=["maple/maple"],
    include_package_data=True,
    long_description=long_description,
    classifiers=[
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: MIT License",
    ],
    install_requires=DEPENDENCIES,
)



"""Backend interface"""

import os

from . import docker
from . import singularity
from . import podman


def Backend():
    """
    Function to call backend based on 'maple_backend' environment variable

    Returns
    -------
    docker/singularity backend module

    """
    backend_dict = {"docker": docker, "singularity": singularity, "podman": podman}

    return backend_dict[os.getenv("maple_backend")]



"""Metadata for PyMaple package"""

__pkgname__ = "PyMaple"
__version__ = "2.0"
__authors__ = "Akash Dhruv"
__license__ = "MIT License"
__copyright__ = "Copyright (c) Akash Dhruv 2021. All Rights Reserved."
__description__ = "Maple Container Utility"



"""Python API for maple"""

import os

from . import Maple
from ..backend import Backend


class Container(Maple):
    """
    Class to manage containers
    """

    def __init__(self, **attributes):
        """
        Parameters
        ----------
        attributes : dictionary
                     { 'name'      : container name,
                       'source'    : source directory,
                       'target'    : target directory,
        """
        self._name = "ubuntu"
        self._source = None
        self._target = None

        super().__init__(self.__dict__, attributes)

        # Condition to check if target and source directories are defined in the Maplefile
        # assign default if they are not, and deal with exceptions
        if not self._target:
            self._target = "/home/mount"
        if not self._source:
            self._source = os.getenv("PWD")

        # Set container name for environment variables
        self._container = self._name

    @property
    def name(self):
        """
        Getter for container name
        """
        return self._name

    def run(self, image, command, options=""):
        """
        Run a container and commit to image
        """
        image.setenv()
        self.setenv()
        Backend().container.run(command, options)



"""Python API for singularity interface in maple"""

import subprocess


def login():
    """
    Login to container account
    """
    print("[maple.system.login] command not available for singularity backend")


def prune():
    """
    Prune system
    """
    subprocess.run("singularity cache clean", shell=True, check=True)



"""Python API for maple"""

import os


class Maple:
    """
    Base class for defining maple environment

    Parameters
    ----------
    default_attributes : dictionary of default attributes
    attributes         : dictionary of user attributes
    """

    def __init__(self, default_attributes, attributes):
        super().__init__()
        self._set_attributes(default_attributes, attributes)

    def setenv(self):
        """
        Set environment variables
        """
        for key, value in self.__dict__.items():
            if value and key != "_name":
                os.environ["maple" + key] = str(value)

    def _set_attributes(self, default_attributes, attributes):
        """
        Private method for initialization

        Parameters
        ----------
        default_attributes : dictionary of default attributes
        attributes         : dictionary of user attributes
        """

        for key in attributes:
            if "_" + key in default_attributes:
                default_attributes["_" + key] = attributes[key]
            else:
                raise ValueError(f'[maple]: attribute "{key}" not present')

        for key, value in default_attributes.items():
            setattr(self, key, value)



"""Python API for singularity interface in maple"""

import os
import random
import subprocess


def commit():
    """
    Commit changes from local container to local image
    """
    print("[maple.container.commit] not available for singularity backend")


def pour(options="--no-home"):
    """
    Pour local image in a container, opposite of maple rinse

    Arguments
    ---------
    options : string of options
    """
    process = subprocess.run(
        f"singularity instance start {options} --bind $maple_source:$maple_target \
                                              $maple_home/images/$maple_image.sif \
                                              $maple_container",
        shell=True,
        check=True,
    )

    if process.returncode != 0:
        raise Exception("[maple] Error inside container")


def rinse(rinse_all=False):
    """
    Stop and remove the local container, opposite of maple pour

    Arguments
    ---------
    rinse_all : (True/False) flag to rinse all container
    """
    if rinse_all:
        subprocess.run("singularity instance stop --all", shell=True, check=True)
    else:
        subprocess.run(
            "singularity instance stop $maple_container", shell=True, check=True
        )


def shell():
    """
    Get shell access to the local container
    """
    subprocess.run(
        "singularity shell --pwd $maple_target instance://$maple_container",
        shell=True,
        check=True,
    )


def run(command, options=""):
    """
    Run and rinse the local container

    Arguments
    ---------
    command : command string
    options : run options
    """
    os.environ["maple_container"] = (
        os.getenv("maple_container") + "_" + str(random.randint(1111, 9999))
    )

    command = f'"{command}"'
    process = subprocess.run(
        f"singularity exec {options} --no-home --bind $maple_source:$maple_target \
                                              --pwd  $maple_target \
                               $maple_home/images/$maple_image.sif bash -c {command}",
        shell=True,
        check=True,
    )

    if process.returncode != 0:
        raise Exception("[maple] Error inside container")


def execute(command):
    """
    Run local image in a container

    Arguments
    ---------
    command : command string
    """
    command = f'"{command}"'
    process = subprocess.run(
        f"singularity exec --pwd $maple_target \
                          instance://$maple_container bash -c {command}",
        shell=True,
        check=True,
    )

    return process.returncode


def publish(cmd_list=None):
    """
    Publish container to an image

    Arguments
    ---------
    cmd_list: list of commands to publish
    """
    print("[maple.container.publish] not available for singularity backend")


def notebook(port="4321"):
    """
    Launch ipython notebook inside the container

    Arguments
    ---------
    port  : port id ('4321')
    """
    os.environ["maple_container"] = (
        os.getenv("maple_container") + "_" + str(random.randint(1111, 9999))
    )

    pour(options="--cleanenv")
    result = execute(f"jupyter notebook --port={port} --no-browser --ip=0.0.0.0")
    rinse()

    if result != 0:
        raise Exception("[maple] Error inside container")


def list():
    """
    List all containers on system
    """
    subprocess.run("singularity instance list", shell=True, check=True)
